# -*- coding: UTF-8 -*-
###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# libsecure.py
#
# classes utilitaires pour lancement des services en https
#
###########################################################################

import os, glob
from subprocess import Popen, PIPE
from creole.parsedico import parse_dico
from creole.fonctionseole import get_module_name
from creole.config import cert_file, key_file
from OpenSSL import SSL

ssl_dir = "/etc/ssl"
cert_dir = os.path.join(ssl_dir, "certs")
key_dir = os.path.join(ssl_dir, "private")
tmp_keyfile = os.path.join(key_dir, "tmpkey.key")
file_serial = os.path.join(ssl_dir, "serial")
req_dir = os.path.join(ssl_dir, "req")
local_ca_dir = os.path.join(ssl_dir, "local_ca")
newcerts_dir = os.path.join(ssl_dir, "newcerts")
ca_conf_file = '/etc/sysconfig/eole/ca-eole.conf'
conf_file = '/etc/sysconfig/eole/certif-eole.conf'

min_serial = int(0x20)

# chargement de la configuration creole
dico = parse_dico()
# recherche d'un éventuel certificat configuré
if dico.get('server_cert',''):
    cert_file = dico['server_cert']
if dico.get('server_key',''):
    key_file = dico['server_key']
# chemin de la CA
ca_file = os.path.join(cert_dir, "ca_local.crt")
ca_dest_file = os.path.join(cert_dir, "ca.crt")
ca_keyfile = os.path.join(key_dir, "ca.key")

def prep_dir() :
    """
    Création de l'arborescence pour openssl
    """
    #on génère le random
    rand_file = os.path.join(ssl_dir, ".rand")
    if not os.path.isfile(rand_file) :
        cmd_random = "/bin/dd if=/dev/urandom of=%s bs=1k count=16 >/dev/null 2>&1" % (rand_file)
        cmd = Popen(cmd_random, shell=True)
        res = cmd.wait()
        if res != 0:
            raise Exception("! erreur lors de la génération du fichier d'entropie !")
    #on crée les fichiers pour gerer la pki
    file_serial = os.path.join(ssl_dir, "serial")
    if not os.path.isfile(file_serial) :
        f = file(file_serial, "w")
        f.write("20")
        f.close()
    file_index = os.path.join(ssl_dir, "index.txt")
    if not os.path.isfile(file_index) :
        f = file(file_index, "w")
        f.close()
    newcerts = os.path.join(ssl_dir, "newcerts")
    if not os.path.isdir(newcerts):
        os.makedirs(newcerts)
    if not os.path.isdir(req_dir):
        os.makedirs(req_dir)
    if not os.path.isdir(local_ca_dir):
        os.makedirs(local_ca_dir)
    ##cmd = Popen("chmod 611 %s" % (key_dir), shell=True)

def sup_passwd(tmp_keyfile, keyfile) :
    """
    Supression de la passphrase sur la clef privée
    """
    key_cmd = "/usr/bin/openssl rsa -in %s -passin pass:secret -out %s >/dev/null 2>&1" % (tmp_keyfile, keyfile)
    cmd = Popen(key_cmd, shell=True)
    res = cmd.wait()
    if res != 0:
        raise Exception('! erreur lors de la génération de la clé ssl dans %s !' % keyfile)
    if os.path.isfile(tmp_keyfile):
        os.unlink(tmp_keyfile)

def finalise_cert (certfile, keyfile, key_user='', key_grp='', key_chmod='',
        cert_user='', cert_grp='', cert_chmod=''):
    """
    Finalisation du certif
    """
    if key_user != '':
        try:
            res = Popen("chown %s %s" % (key_user, keyfile), shell=True).wait()
            assert res == 0
        except:
            print "\n! Impossible de changer les droits de %s" % keyfile
            return False
    if key_grp != '':
        try:
            res=Popen("/bin/chgrp %s %s" % (key_grp, keyfile), shell=True).wait()
            assert res == 0
        except:
            print "\n! Impossible de changer les droits de %s" % keyfile
            return False
    if key_chmod != '':
        try:
            res = Popen("/bin/chmod %s %s" % (key_chmod, keyfile), shell=True).wait()
            assert res == 0
        except:
            print "\n! Impossible de changer les droits de %s" % keyfile
            return False
    if cert_user != '':
        try:
            res = Popen("/bin/chown %s %s" % (cert_user, certfile), shell=True).wait()
            assert res == 0
        except:
            print "\n! Impossible de changer les droits de %s" % certfile
            return False
    if cert_grp != '':
        try:
            res = Popen("/bin/chgrp %s %s" % (cert_grp, certfile), shell=True).wait()
            assert res == 0
        except:
            print "\n! Impossible de changer les droits de %s" % certfile
            return False
    if cert_chmod != '':
        try:
            res = Popen("/bin/chmod %s %s" % (cert_chmod, certfile), shell=True).wait()
            assert res == 0
        except:
            print "\n! Impossible de changer les droits de %s" % certfile
            return False
    return True

def merge_fic(dst_fic, in_fics, overwrite=False):
    """
    Merge d'une liste de fichiers dans un fichier de destination
    (le contenu d'origine est conservé)
    """
    if type(in_fics) != list:
        in_fics = [in_fics]
    for fic in in_fics:
        if not os.path.isfile(fic):
            print "Erreur le fichier %s n'existe pas" % fic
    data = ""
    for fic_src in in_fics:
        f_src = file(fic_src)
        data += f_src.read().rstrip() + '\n'
        f_src.close()
    if overwrite:
        f_dst = file(dst_fic, "w")
    else:
        f_dst = file(dst_fic, "a+")
    f_dst.write(data)
    f_dst.close()

def gen_certs(regen=False):
    """
    Génère la ca puis les certificats
    """
    verif_ca()
    ca_generated = gen_ca(regen)
    merge_ca()
    module = get_module_name()
    if ca_generated:
        regen = True
    gen_default_certs(module, regen=regen)

def verif_ca():
    """
    vérifie que la ca est générée correctement (serial > 0x20) et cn valide
    """
    # gestion des anciennes version de ca.crt
    if os.path.isfile(ca_dest_file) and not os.path.isfile(ca_file):
        # on reprend le premier certificat présent dans ca.crt dans ca_local.crt
        ca_certs = open(ca_dest_file).read().strip()
        tag_begin = '-----BEGIN CERTIFICATE-----'
        try:
            ca_data = tag_begin + ca_certs.split(tag_begin)[1]
            local_ca = open(ca_file, 'w')
            local_ca.write(ca_data)
            local_ca.close()
        except IndexError:
            # impossible de reprendre la ca actuelle, elle sera regénérée
            pass
    serial = int(0x20)
    # vérification de la valeur actuelle du ca
    # vérification du cn de la ca
    if os.path.isfile(ca_file):
        cmd = Popen(['/usr/bin/openssl', 'x509', '-in', ca_file, '-subject', '-noout'], stdout=PIPE)
        if cmd.wait() != 0:
            os.unlink(ca_file)
            prep_dir()
        else:
            res = cmd.communicate()[0]
            cn = res[res.rindex('CN=')+3:res.rindex('CN=')+15].strip()
            if cn == str('self signed ') or cn == str('CA auto sign'):
                os.unlink(ca_file)
                prep_dir()
    if os.path.isfile(file_serial):
        serial = open(file_serial).read().strip()
        # conversion en hexa
        serial = int(serial, 16)
        if serial < min_serial:
            if os.path.isfile(ca_file):
                os.unlink(ca_file)
            os.unlink(file_serial)
            for f_index in glob.glob(os.path.join(ssl_dir, 'index*')):
                os.unlink(f_index)
            for f_cert in glob.glob(os.path.join(newcerts_dir, '*.pem')):
                os.unlink(f_cert)
            prep_dir()

def gen_ca(regen=False):
    """
    Generation ca
    """
    generated = False
    prep_dir()
    old_keyfile = os.path.join(cert_dir, "ca.key")
    if not os.path.isfile(ca_conf_file):
        raise Exception("Fichier template de la configuration du certificat non trouvé :\n\t%s\n" % ca_conf_file)
    # déplacement de l'ancienne clé si nécessaire
    if os.path.isfile(old_keyfile) and not os.path.isfile(ca_keyfile):
        try:
            os.rename(old_keyfile, ca_keyfile)
            print("Le fichier %s (clé de la CA) a été déplacé dans %s" % (old_keyfile, key_dir))
        except:
            raise Exception("La clé de la CA actuelle n'a pas pu être recopiée vers %s") % ca_keyfile
    if regen or (not os.path.isfile(ca_keyfile)) or (not os.path.isfile(ca_file)):
        print("Génération du certificat de la CA")
        ## On genère le certif de l'ac
        ca_gen = "/usr/bin/openssl req -x509 -config %s -newkey rsa:%s -days %s -keyout %s -out %s >/dev/null 2>&1" % (ca_conf_file, dico['ssl_default_key_bits'], dico['ssl_default_cert_time'], tmp_keyfile, ca_file)
        cmd = Popen(ca_gen, shell=True)
        if cmd.wait() != 0:
            raise Exception("Erreur lors de la génération de la CA")
        sup_passwd(tmp_keyfile, ca_keyfile)
        generated = True
    ## génération d'une crl
    if not os.path.isfile(os.path.join(ssl_dir, 'eole.crl')):
        print("Génération de la liste de révocation")
        crl_gen = "/usr/bin/openssl ca -gencrl -config %s -crldays %s -out %s/eole.crl >/dev/null 2>&1" % (ca_conf_file, dico['ssl_default_cert_time'], ssl_dir)
        cmd = Popen(crl_gen, shell=True)
        if cmd.wait() != 0:
            raise Exception("Erreur lors de la génération de la CRL (%s/eole.crl)" % ssl_dir)
    ## application des droits
    finalise_cert(ca_file, ca_keyfile, key_chmod='600')
    return generated

def merge_ca():
    """
    concatène toutes les ca utiles dans ca.crt
    """
    ## concaténation des certificats education
    ca_list = [ca_file, os.path.join(cert_dir, 'ACInfraEducation.pem')]
    ## concaténation de certificats supplémentaires si définis
    for ca_perso in glob.glob(os.path.join(local_ca_dir,'*.*')):
        if os.path.isfile(ca_perso):
            ca_list.append(ca_perso)
    merge_fic(ca_dest_file, ca_list, True)

def gen_certif(certfile=None, keyfile=None, key_user='', key_grp='', key_chmod='',
        cert_user='', cert_grp='', cert_chmod='', regen=False, copy_key=False):
    """
    Génération des requêtes de certificats et signature par la CA
    """
    if not os.path.isfile(conf_file):
        raise Exception("Fichier template de la configuration du certificat non trouvé :\n\t%s\n" % conf_file)

    if certfile is None:
        # generation du certif par defaut (eole.crt)
        certfile = cert_file
        keyfile = key_file
        basefile = os.path.splitext(certfile)[0]
    else:
        basefile = os.path.splitext(certfile)[0]
        if keyfile is None:
            keyfile = "%s.key" % (basefile)
    if regen or not os.path.isfile(certfile) or not os.path.isfile(keyfile):
        # certificat absent ou regénération demandée
        fic_p10 = os.path.join(req_dir, "%s.p10" % (os.path.basename(basefile)))
        # génération de la requête de certificat x509 et d'un simili certificat auto-signé
        gen_req = "/usr/bin/openssl req -new -newkey rsa:%s -days %s -config %s -keyout %s -out %s >/dev/null 2>&1" % (
                  dico['ssl_default_key_bits'], dico['ssl_default_cert_time'], conf_file, tmp_keyfile, fic_p10)
        cmd = Popen(gen_req, shell=True)
        if cmd.wait() != 0:
            raise Exception('! Erreur lors de la génération de la requête ssl dans %s !' % fic_p10)
        sup_passwd(tmp_keyfile, keyfile)
        # on signe la requête
        ca_signe = "/usr/bin/openssl ca -in %s -config %s -out %s -batch -notext >/dev/null 2>&1" % (fic_p10, ca_conf_file, certfile)
        cmd = Popen(ca_signe, shell=True)
        if cmd.wait() != 0:
            raise Exception('! Erreur lors de la signature de la requête dans %s !' % fic_p10)
        print("* Certificat %s généré" % certfile)
        if copy_key:
            merge_fic(certfile, [keyfile])
    finalise_cert(certfile, keyfile, key_user=key_user,
                            key_grp=key_grp, key_chmod=key_chmod,
                            cert_user=cert_user, cert_grp=cert_grp, cert_chmod=cert_chmod)
    return True

def gen_default_certs(module, regen=False):
    """
    Procédure de génération du jeu de certificats pour les modules Eole
    """
    gen_certif(regen=regen)
    if cert_file != os.path.join(cert_dir, 'eole.crt'):
        # on génère le certificat eole.crt dans tous les cas
        gen_certif(os.path.join(cert_dir, 'eole.crt'), os.path.join(cert_dir, 'eole.key'), regen=regen)
    # création de eole.pem (concaténation de eole.crt et eole.key)
    merge_fic(os.path.join(cert_dir, 'eole.pem'), [os.path.join(cert_dir, 'eole.crt'), os.path.join(cert_dir, 'eole.key')], True)
    if module == 'scribe':
        gen_certif('/etc/ssl/certs/scribe.crt', regen=regen)
        gen_certif('/etc/courier/imapd.pem', cert_user='root', cert_grp='root', cert_chmod='600',
                   key_user='root', key_grp='root', key_chmod='600', copy_key=True, regen=regen)
        gen_certif('/etc/courier/pop3d.pem', cert_user='root', cert_grp='root', cert_chmod='600',
                   key_user='root', key_grp='root', key_chmod='600', copy_key=True, regen=regen)
    if module == 'amon' or module == 'amonecole':
        # certificats NuFw
        gen_certif('/etc/nufw/certs/nufw-eole.crt', cert_user='root', cert_grp='root', cert_chmod='400', regen=regen)
        # certificats NuAuth
        # droits gérés dans reconfigure
        gen_certif('/etc/nufw/certs/nuauth-eole.crt', regen=regen)
        gen_dh = "/usr/bin/openssl dhparam -out %s 1024 >/dev/null 2>&1" % (os.path.join(ssl_dir, "dh"))
        Popen(gen_dh, shell=True)

class ServerContextFactory:
    """
    Factory permettant de créer un contexte SSL
    à utiliser pour les serveurs twisted (xmlrpc ...)
    """

    def __init__(self, certfile='', keyfile='', cafile=''):
        self.keyfile = None
        if certfile != '':
            self.certfile = certfile
        else:
            self.certfile = cert_file
        if keyfile != '':
            self.keyfile = keyfile
        else:
            self.keyfile = key_file
        if cafile != '':
            self.cafile = cafile
        else:
            self.cafile = ca_dest_file

    def getContext(self):
        """Initialisation du contexte SSL et chargement des certificats
        """
        ctx = SSL.Context(SSL.SSLv23_METHOD)
        ctx.use_certificate_file(self.certfile)
        try:
            ctx.use_privatekey_file(self.certfile)
        except:
            if self.keyfile is not None:
                alt_keyfile = self.keyfile
            else:
                # impossible de charger la clé depuis le fichier du certificat on regarde si il n'y a pas un fichier .key associé
                alt_keyfile = os.path.splitext(self.certfile)[0] + '.key'
            ctx.use_privatekey_file(alt_keyfile)
        if self.cafile and os.path.isfile(self.cafile):
            try:
                # chargement des ca à envoyer au client
                ctx.load_client_ca(self.cafile)
            except Exception, e:
                # En cas d'erreur, on continue quand même.
                print 'Erreur au chargement des CA à envoyer aux clients (%s) : %s' % (self.cafile, str(e))
        return ctx
